package external

object Components {
  // Base Component
  trait Expr[-V[-_,_]] {
    def accept[a](vis : V[Expr[V],a]) : a
  }
  
  // Numeric components
  trait vnum[-R,A] {
    def num(x : Int) : A 
  }
  
  case class Num[-V[-R,A] <: vnum[R,A]](x : Int) extends Expr[V] {
    def accept[a](vis : V[Expr[V],a]) : a = vis.num(x)
  }
  
  // Addition components
  trait vadd[-R,A] {
    def add(e1 : R, e2 : R) : A
  }
  
  case class Add[-V[-R,A] <: vadd[R,A]](e1 : Expr[V], e2 : Expr[V]) extends Expr[V] {
    def accept[a](vis : V[Expr[V],a]) : a = vis.add(e1,e2)
  }
  
  // Negation components
  trait vneg[-R,A] {
    def neg(e : R) : A
  }
    
  case class Neg[-V[-R,A] <: vneg[R,A]](e : Expr[V]) extends Expr[V] {
    def accept[a](vis : V[Expr[V],a]) : a = vis.neg(e)
  }
  
  // Evaluation component,
  trait BaseEval[V[-_,_]] extends vnum[Expr[V],Int] with vadd[Expr[V],Int] with vneg[Expr[V],Int] {
    self : V[Expr[V],Int] =>
    
    def num(x : Int)                     = x
    def add(e1 : Expr[V], e2 : Expr[V])  = e1.accept(this) + e2.accept(this) 
    def neg(e : Expr[V])                 = -e.accept(this)
  }
  
  // Pretty printing component
  trait BasePrint[V[-_,_]] extends vnum[Expr[V],String] with vadd[Expr[V],String] with vneg[Expr[V],String] {
    self : V[Expr[V],String] =>
    
    def num(x : Int)                     = x.toString()
    def add(e1 : Expr[V], e2 : Expr[V])  = "(" + e1.accept(this) + " + " + e2.accept(this) + ")"
    def neg(e : Expr[V])                 = "-" + e.accept(this)
  }
  
  // Components for the narrow operation
  trait NumNarrow[V1[-_,_],V2[-R,X] <: vnum[R,X]] extends vnum[Expr[V1],Expr[V2]] {
    def num(x : Int)                     = Num[V2](x)
  }
  
  trait AddNarrow[V1[-_,_],V2[-R,X] <: vadd[R,X]] extends vadd[Expr[V1],Expr[V2]] {
    self : V1[Expr[V1],Expr[V2]] =>
    
    def add(e1 : Expr[V1], e2 : Expr[V1])  = Add[V2](e1.accept(this),e2.accept(this))
  }
  
  // Components needed for subtraction
  trait vminus[-R,A] {
    def minus(e1 : R, e2 : R) : A
  }
  
  case class Minus[-V[-R,X] <: vminus[R,X]](e1 : Expr[V], e2 : Expr[V]) extends Expr[V] {
    def accept[a](vis : V[Expr[V],a]) : a = vis.minus(e1,e2)
  }
  
  // More narrow components
  trait NegNarrow[V1[-_,_],V2[-R,X] <: vneg[R,X]] extends vneg[Expr[V1],Expr[V2]] {
    self : V1[Expr[V1],Expr[V2]] =>
    
    def neg(e : Expr[V1]) = Neg[V2](e.accept(this))
  }
  
  trait NMNarrow[V1[-_,_],V2[-R,X] <: vnum[R,X] with vminus[R,X]] extends vneg[Expr[V1],Expr[V2]] {
    self : V1[Expr[V1],Expr[V2]] =>
    
    def neg(e : Expr[V1])                 = Minus[V2](Num[V2](0),e.accept(this))
  }
}